#include <stdio.h>
#include <stdlib.h>

/* matMul-mpi.c: matrix multiplication in 1D representation of matrices */
/*               (MPI version)                              */

/*
 * Calculate C = A * B
 * A, B, C: 1D-array representation of matrices (size x strd).
 * Thus, the (i,j) element is accessed by A[i * strd + j], etc.
 */

#include <mpi.h>

void matMul(double *matA, double *matB, double *matC,
	double *matAl, double *matCl,
	int size, int strd,  int myid, int nprocs) {
    /* local variables */
    int i,j,k;
    int rows;

    /* initialize */
    rows = size / nprocs;
    MPI_Scatter( matA, rows*strd, MPI_DOUBLE,
	         matAl, rows*strd, MPI_DOUBLE,
		 0, MPI_COMM_WORLD );
    MPI_Bcast( matB, size*strd, MPI_DOUBLE, 0, MPI_COMM_WORLD );
    for (i=0; i<rows; i++) {
        for (j=0; j<size; j++) {
            matCl[i * strd + j] = 0.e0;
        }
    }

    /* matrix multiplication */
    for (i=0; i<rows; i++) {
        for (j=0; j<size; j++) {
            for (k=0; k<size; k++) {
                matCl[i * strd + j] +=
		    matAl[i * strd + k]*matB[k * strd + j];
            }
        }
    }
    MPI_Gather(matCl, rows*strd, MPI_DOUBLE,
	       matC, rows*strd, MPI_DOUBLE,
	       0, MPI_COMM_WORLD );
}
