#include <stdio.h>
#include <stdlib.h>

/* matMul2d-mpi.c: matrix multiplication in 2D representation of matrices */
/*                 (MPI version)                                          */

/*
 * Calculate C = A * B
 * A, B, C: 2D-array representation of matrices (size x strd).
 * Thus, the (i,j) element is accessed by A[i][j], etc.
 */

#include <mpi.h>

void matMul(double **matA, double **matB, double **matC,
	double **matAl, double **matCl,
	int size, int strd,  int myid, int nprocs) {
    /* local variables */
    int i,j,k;
    int rows;

    rows = size / nprocs;
    MPI_Scatter( &matA[0][0], rows*strd, MPI_DOUBLE,
	         &matAl[0][0], rows*strd, MPI_DOUBLE,
		 0, MPI_COMM_WORLD );
    MPI_Bcast( &matB[0][0], size*strd, MPI_DOUBLE, 0, MPI_COMM_WORLD );
    /* initialize */
    for (i=0; i<rows; i++) {
        for (j=0; j<size; j++) {
            matCl[i][j] = 0.e0;
        }
    }

    /* matrix multiplication */
    for (i=0; i<rows; i++) {
        for (j=0; j<size; j++) {
            for (k=0; k<size; k++) {
                matCl[i][j] += matAl[i][k]*matB[k][j];
            }
        }
    }
    MPI_Gather(&matCl[0][0], rows*strd, MPI_DOUBLE,
	       &matC[0][0], rows*strd, MPI_DOUBLE,
	       0, MPI_COMM_WORLD );
}
