#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <mpi.h>

extern double second();
extern double** matAlloc(int size, int strd);
extern void matFree(double **mat);

int main(int argc, char* argv[]) {
    int size[] = {16, 32, 64, 128, 256, 512, 1024, 2048};
    int i, j, nitera, n, Nsize=8, sz, it, strd;
    double **matA, **matB, **matC;
    double **matAl, **matCl;
    double coef, pi, fact, err, diff, R;
    double t0, t1, gflops;
    double tmin=10.e0;
    int myid, nprocs;
    int rows;

    MPI_Init( &argc, &argv );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs);
    MPI_Comm_rank( MPI_COMM_WORLD, &myid);
    /* initialize */
    pi = 4.e0*atan(1.e0);
    if (Nsize > 8) Nsize=8;

    if (myid == 0) {
	printf("\n\n Matrix multiplication in 2D-representation\n");
	printf(" (MPI version)  nprocs = %d\n", nprocs);
	printf(" CPU time is measured.\n");
	printf(" Minimum duration for a measurement is 10 (sec).\n\n");
	printf(" size   time(sec)    n Calc     Gflops      ERROR\n");
	printf("----- ------------ ---------- ----------- -------------\n");
    }
    for (sz=0; sz<Nsize; sz++) {
	n = size[sz];
	strd = n;
	coef = sqrt(2.e0/n);
	fact = pi/n;
	rows = n / nprocs;
        if (rows < 1 ) continue;
	matA = matAlloc(n,strd);
	matB = matAlloc(n,strd);
	matC = matAlloc(n,strd);
	matAl = matAlloc(rows, strd);
	matCl = matAlloc(rows, strd);
	if (myid ==0 ) {
	    for (i=0; i<n; i++) {
		for (j=0; j<n; j++) {
		    matA[i][j] = coef * sin( fact*(i-j) );
		    matB[i][j] = coef * cos( fact*(i-j) );
		}
	    }
	}
	nitera=it=1;
	t0 = second();
	while (1) {
	    matMul(matA, matB, matC, matAl, matCl, n, strd, myid, nprocs);
	    t1 = second();
	    if ( it >= nitera ) {
		if ( (t1-t0) >= tmin ) {
		    break;
		} else {
		    nitera *= 2;
		}
	    }
	    it++;
	}
	/* check result */
	if (myid == 0) {
	    gflops = nitera*2.e0*n*n*n/(t1-t0)*1.e-9;
	    err = 0.e0;
	    for (i=0; i<n; i++) {
		for (j=0; j<n; j++) {
		    diff = sin( fact*(i-j) ) - matC[i][j];
		    err += (diff*diff);
		}
	    }
	    R = sqrt(err);
	    printf("%5d  %10.5f  %9d  %10.6f  %12.5e\n",
		    n, (t1-t0), nitera, gflops, R);
	}
	matFree(matA);
	matFree(matB);
	matFree(matC);
	matFree(matAl);
	matFree(matCl);
    }
    MPI_Finalize();
    return 0;
}
